import abc
import typing
import anchor.builtins as builtins


__all__: list[str] = ['AST', 'SYMTABLE',]


class Factory(abc.ABC):

    def __init__(self):
        self._builders: dict[str, typing.Any] = {}

    def _registerbuilder(self, key: str, builder: typing.Any) -> None:
        self._builders[key] = builder

    def new(self, key, **kwargs) -> typing.Any:
        builder: typing.Any = self._builders.get(key)
        if (not builder):
            raise ValueError(key)
        return builder(**kwargs)


class ASTNodeFactory(object):

    class __ASTNodeFactory(Factory):

        __astnodename: dict[typing.Type, str] = dict({
            bool: 'Boolean',
            type(None): 'Null',
            int: 'Integer',
            float: 'Float',
            complex: 'Complex',
            str: 'String',
            tuple: 'Tuple',
            list: 'List',
            dict: 'Dict',
        })

        def __init__(self, declarations: list[tuple[str, typing.Type]]):
            super().__init__()
            for key, builder in declarations:
                self._registerbuilder(key, builder)
        
        def new(self, value: typing.Any) -> typing.Any:
            if (isinstance(value, builtins.Type)):
                return super().new(value.typename, value=value)
            key = self.__astnodename[type(value)]
            return super().new(key, value=value)

    __instance: __ASTNodeFactory = None

    def __new__(cls, declarations: list[tuple[str, typing.Type]]):
        if (not ASTNodeFactory.__instance):
            ASTNodeFactory.__instance = \
                ASTNodeFactory.__ASTNodeFactory(declarations)
        return ASTNodeFactory.__instance

    def new(self, value: typing.Any) -> typing.Any:
        return ASTNodeFactory().new(value)


class SymbolTableFactory(object):

    class __SymbolTableFactory(Factory):

        def __init__(self, declarations: list[tuple[str, typing.Type]]):
            super().__init__()
            for key, builder in declarations:
                self._registerbuilder(key, builder)

    __instance: __SymbolTableFactory = None

    def __new__(cls, declarations: list[tuple[str, typing.Type]]):
        if (not SymbolTableFactory.__instance):
            SymbolTableFactory.__instance = \
                SymbolTableFactory.__SymbolTableFactory(declarations)
        return SymbolTableFactory.__instance

    def new(self, key: str, **kwargs) -> typing.Any:
        return SymbolTableFactory().new(key, **kwargs)


AST: ASTNodeFactory = None
SYMTABLE: SymbolTableFactory = None



import abc
import anchor.ply.yacc as yacc
import anchor.lex as lex
import anchor.token as token
import anchor.ast as ast
import anchor.system as system


__all__: list[str] = ['AnchorParser',]


class Parser(abc.ABC):

    tokens: tuple = ()
    precedence: tuple = ()

    def __init__(self, **kwargs):
        self.debuglex: bool = kwargs.get('debuglex', False)
        self.debugyacc: bool = kwargs.get('debugyacc', False)
        self.debuglog: bool = kwargs.get('debuglog', None)

        # Build the lexer and parser
        self.lexer: lex.AnchorLexer = lex.AnchorLexer()
        self.lexer.build(
            debug=self.debuglex, 
            debuglog=self.debuglog if self.debuglex else None,
        )
        self.parser = yacc.yacc(
            module=self,
            debug=self.debugyacc,
            debuglog=self.debuglog if self.debugyacc else None,
        )

    def parse(self, data: str) -> ast.ASTNode:
        if (self.debuglex): self.lexer.debug(data)
        return self.parser.parse(data)


class AnchorParser(Parser):

    # Tokens list
    tokens: tuple = lex.AnchorLexer.tokens

    # Precedence rules
    precedence: tuple = (
        ('left', token.OR,),
        ('left', token.AND,),
        ('right', token.NOT,),
        ('nonassoc', token.EQEQUAL, token.NOTEQUAL, 
            token.LESS, token.GREATER, token.LESSEQUAL, token.GREATEREQUAL,),
        ('left', token.PLUS, token.MINUS,),
        ('left', token.STAR, token.SLASH, token.DOUBLESLASH, token.PERCENT,),
        ('right', token.UPLUS, token.UMINUS,),
        ('left', token.DOUBLESTAR,),
        ('left', token.LPAR, token.RPAR,),
    )

    def p_program(self, p: yacc.YaccProduction) -> None:
        '''program : block
                   | empty'''
        if (p[1]):
            block: ast.Block = p[1]
            p[0] = ast.Program(block)

    def p_block(self, p: yacc.YaccProduction) -> None:
        '''block : statements'''
        statements: list[ast.Statement] = p[1]
        p[0] = ast.Block(statements)

    def p_statements(self, p: yacc.YaccProduction) -> None:
        '''statements : statements statement
                      | statement'''
        if (len(p) == 3):
            statements: list[ast.Statement] = p[1]
            statement: ast.Statement = p[2]
            if (statement): statements.append(statement)
            p[0] = statements
        elif (len(p) == 2):
            statement: ast.Statement = p[1]
            p[0] = list([statement])

    def p_statement_assignment(self, p: yacc.YaccProduction) -> None:
        '''statement : name EQUAL expression SEMI'''
        name: ast.Name = p[1]
        expression: ast.Expression = p[3]
        p[0] = ast.Assignment(name, expression)

    def p_statement_if(self, p: yacc.YaccProduction) -> None:
        '''statement : IF expression THEN block elifs END'''
        expression: ast.Expression = p[2]
        block: ast.Block = p[4]
        elifs: list[ast.Elif] = p[5]
        p[0] = ast.If(expression, block, elifs=elifs)

    def p_statement_ifelse(self, p: yacc.YaccProduction) -> None:
        '''statement : IF expression THEN block elseblock END
                     | IF expression THEN block END'''
        if (len(p) == 7):
            expression: ast.Expression = p[2]
            block: ast.Block = p[4]
            elseblock: ast.Block = p[5]
            p[0] = ast.If(expression, block, elseblock=elseblock)
        elif (len(p) == 6):
            expression: ast.Expression = p[2]
            block: ast.Block = p[4]
            p[0] = ast.If(expression, block)

    def p_elifs(self, p: yacc.YaccProduction) -> None:
        '''elifs : ELIF expression THEN block elifs'''
        expression: ast.Expression = p[2]
        block: ast.Block = p[4]
        elifs: list[ast.Elif] = p[5]
        p[0] = list([ast.Elif(expression, block)]) + elifs

    def p_elifs_else(self, p: yacc.YaccProduction) -> None:
        '''elifs : ELIF expression THEN block elseblock
                 | ELIF expression THEN block'''
        if (len(p) == 6):
            expression: ast.Expression = p[2]
            block: ast.Block = p[4]
            elseblock: ast.Block = p[5]
            p[0] = list([ast.Elif(expression, block, elseblock)])
        elif (len(p) == 5):
            expression: ast.Expression = p[2]
            block: ast.Block = p[4]
            p[0] = list([ast.Elif(expression, block)])

    def p_elseblock(self, p: yacc.YaccProduction) -> None:
        '''elseblock : ELSE block'''
        block: ast.Block = p[2]
        p[0] = block

    def p_statement_iterate(self, p: yacc.YaccProduction) -> None:
        '''statement : ITERATE expression FOR name BEGIN block END'''
        iterable: ast.Expression = p[2]
        variable: ast.Name = p[4]
        block: ast.Block = p[6]
        p[0] = ast.Iterate(iterable, variable, block)

    def p_statement_loop(self, p: yacc.YaccProduction) -> None:
        '''statement : LOOP expression BEGIN block END'''
        expression: ast.Expression = p[2]
        block: ast.Block = p[4]
        p[0] = ast.Loop(expression, block)

    def p_statement_functiondef(self, p: yacc.YaccProduction) -> None:
        '''statement : FUNCTION name LPAR parameters RPAR \
                       RARROW expression BEGIN block END
                     | FUNCTION name LPAR RPAR \
                       RARROW expression BEGIN block END'''
        if (len(p) == 11):
            name: ast.Name = p[2]
            parameters: list[ast.Parameter] = p[4]
            returntype: ast.Expression = p[7]
            body: ast.Block = p[9]
            p[0] = ast.FunctionDef(
                name, parameters, body, returntype=returntype,
            )
        elif (len(p) == 10):
            name: ast.Name = p[2]
            parameters: list[ast.Parameter] = list()
            returntype: ast.Expression = p[7]
            body: ast.Block = p[8]
            p[0] = ast.FunctionDef(
                name, parameters, body, returntype=returntype,
            )

    def p_statement_classdef(self, p: yacc.YaccProduction) -> None:
        '''statement : CLASS LSQB annotations RSQB name BEGIN block END'''
        annotations: list[ast.Annotation] = p[3]
        name: ast.Name = p[5]
        block: ast.Block = p[7]
        p[0] = ast.ClassDef(name, block, annotations)

    def p_statement_property(self, p: yacc.YaccProduction) -> None:
        '''statement : PROPERTY LSQB annotations RSQB name SEMI'''
        annotations: list[ast.Annotation] = p[3]
        name: ast.Name = p[5]
        p[0] = ast.Property(name, annotations)

    def p_statement_methoddef(self, p: yacc.YaccProduction) -> None:
        '''statement : METHOD LSQB annotations RSQB name LPAR parameters RPAR \
                       RARROW expression BEGIN block END
                     | METHOD LSQB annotations RSQB name LPAR RPAR \
                       RARROW expression BEGIN block END'''
        if (len(p) == 14):
            annotations: list[ast.Annotation] = [p[3]]
            name: ast.Name = p[5]
            parameters: list[ast.Parameter] = p[7]
            returntype: ast.Expression = p[10]
            body: ast.Block = p[12]
            p[0] = ast.MethodDef(
                name, parameters, body, annotations, returntype=returntype,
            )
        elif (len(p) == 13):
            annotations: list[ast.Annotation] = [p[3]]
            name: ast.Name = p[5]
            parameters: list[ast.Parameter] = []
            returntype: ast.Expression = p[9]
            body: ast.Block = p[11]
            p[0] = ast.MethodDef(
                name, parameters, body, annotations, returntype=returntype,
            )

    def p_annotations(self, p: yacc.YaccProduction) -> None:
        '''annotations : _annotations COMMA
                       | _annotations'''
        p[0] = p[1] 

    def p__annotations(self, p: yacc.YaccProduction) -> None:
        '''_annotations : _annotations COMMA annotation
                        | annotation'''
        if (len(p) == 4):
            annotations: list[ast.Annotation] = p[1]
            annotation: ast.Annotation = p[3]
            if (annotation): annotations.append(annotation)
            p[0] = annotations
        elif (len(p) == 2):
            annotation: ast.Annotation = p[1]
            p[0] = list([annotation])

    def p_annotation(self, p: yacc.YaccProduction) -> None:
        '''annotation : PUBLIC
                      | PRIVATE
                      | PROTECTED
                      | FACTORY
                      | GET
                      | SET
                      | REF
                      | VAL'''
        literal: str = p[1]
        p[0] = ast.Annotation(literal)

    def p_parameters(self, p: yacc.YaccProduction) -> None:
        '''parameters : _parameters COMMA
                      | _parameters'''
        p[0] = p[1] 

    def p__parameters(self, p: yacc.YaccProduction) -> None:
        '''_parameters : _parameters COMMA parameter
                       | parameter'''
        if (len(p) == 4):
            parameters: list[ast.Parameter] = p[1]
            parameter: ast.Parameter = p[3]
            if (parameter): parameters.append(parameter)
            p[0] = parameters
        if (len(p) == 2):
            parameter: ast.Parameter = p[1]
            p[0] = list([parameter])

    def p_parameter(self, p: yacc.YaccProduction) -> None:
        '''parameter : name COLON name LSQB annotations RSQB
                     | name COLON name
                     | name'''
        if (len(p) == 7):
            name: ast.Name = p[1]
            typename: ast.Name = p[3]
            p[0] = ast.Parameter(name, typename)
        elif (len(p) == 4):
            name: ast.Name = p[1]
            typename: ast.Name = p[3]
            p[0] = ast.Parameter(name, typename)
        elif (len(p) == 2):
            name: ast.Name = p[1]
            p[0] = ast.Parameter(name)

    def p_statement_break(self, p: yacc.YaccProduction) -> None:
        '''statement : BREAK SEMI'''
        p[0] = ast.Break(literal=p[1])

    def p_statement_continue(self, p: yacc.YaccProduction) -> None:
        '''statement : CONTINUE SEMI'''
        p[0] = ast.Continue(literal=p[1])

    def p_statement_return(self, p: yacc.YaccProduction) -> None:
        '''statement : RETURN expression SEMI'''
        expression = p[2]
        p[0] = ast.Return(expression=expression)

    def p_statement_expression(self, p: yacc.YaccProduction) -> None:
        '''statement : expression SEMI'''
        p[0] = p[1]

    def p_expression_or(self, p: yacc.YaccProduction) -> None:
        '''expression : expression OR expression'''
        p[0] = ast.Or(p[1], p[3])

    def p_expression_and(self, p: yacc.YaccProduction) -> None:
        '''expression : expression AND expression'''
        p[0] = ast.And(p[1], p[3])

    def p_exrepssion_not(self, p: yacc.YaccProduction) -> None:
        '''expression : NOT expression'''
        p[0] = ast.Not(p[2])

    def p_expression_relationalop(self, p: yacc.YaccProduction) -> None:
        '''expression : expression EQEQUAL expression
                      | expression NOTEQUAL expression
                      | expression LESS expression
                      | expression LESSEQUAL expression
                      | expression GREATER expression
                      | expression GREATEREQUAL expression'''
        left: ast.Expression = p[1]
        right: ast.Expression = p[3]
        operator: str = token.NAME[p[2]]
        if (operator == token.EQEQUAL):
            p[0] = ast.EqEqual(left, right)
        elif (operator == token.NOTEQUAL):
            p[0] = ast.NotEqual(left, right)
        elif (operator == token.LESS):
            p[0] = ast.Less(left, right)
        elif (operator == token.LESSEQUAL):
            p[0] = ast.LessEqual(left, right)
        elif (operator == token.GREATER):
            p[0] = ast.Greater(left, right)
        elif (operator == token.GREATEREQUAL):
            p[0] = ast.GreaterEqual(left, right)
    
    def p_expression_binaryop(self, p: yacc.YaccProduction) -> None:
        '''expression : expression PLUS expression
                      | expression MINUS expression
                      | expression STAR expression
                      | expression DOUBLESTAR expression
                      | expression SLASH expression
                      | expression DOUBLESLASH expression
                      | expression PERCENT expression'''
        left: ast.Expression = p[1]
        right: ast.Expression = p[3]
        operator: str = token.NAME[p[2]]
        if (operator == token.PLUS):
            p[0] = ast.Plus(left, right)
        elif (operator == token.MINUS):
            p[0] = ast.Minus(left, right)
        elif (operator == token.STAR):
            p[0] = ast.Star(left, right)
        elif (operator == token.DOUBLESTAR):
            p[0] = ast.DoubleStar(left, right)
        elif (operator == token.SLASH):
            p[0] = ast.Slash(left, right)
        elif (operator == token.DOUBLESLASH):
            p[0] = ast.DoubleSlash(left, right)
        elif (operator == token.PERCENT):
            p[0] = ast.Percent(left, right)

    def p_expression_unaryop(self, p: yacc.YaccProduction) -> None:
        '''expression : PLUS expression %prec UPLUS
                      | MINUS expression %prec UMINUS'''
        right: ast.Expression = p[2]
        operator: str = token.NAME[p[1]]
        if (operator == token.PLUS):
            p[0] = ast.UPlus(right)
        elif (operator == token.MINUS):
            p[0] = ast.UMinus(right)

    def p_expression_group(self, p: yacc.YaccProduction) -> None:
        '''expression : LPAR expression RPAR'''
        p[0] = p[2]

    def p_expression_dotname(self, p: yacc.YaccProduction) -> None:
        '''expression : expression DOT name'''
        expression = p[1]
        name = p[3]
        p[0] = ast.DotName(expression, name)

    def p_expression_call(self, p: yacc.YaccProduction) -> None:
        '''expression : expression LPAR arguments RPAR
                      | expression LPAR RPAR'''
        if (len(p) == 5):
            expression: ast.Expression = p[1]
            arguments: list[ast.Expression] = p[3]
            p[0] = ast.Call(expression, arguments)
        elif (len(p) == 4):
            expression: ast.Expression = p[1]
            arguments: list[ast.Expression] = []
            p[0] = ast.Call(expression, arguments)

    def p_arguments(self, p: yacc.YaccProduction) -> None:
        '''arguments : args COMMA
                     | args'''
        p[0] = p[1] 

    def p_args(self, p: yacc.YaccProduction) -> None:
        '''args : args COMMA expression
                | expression'''
        if (len(p) == 4):
            args: list[ast.Expression] = p[1]
            expression: ast.Expression = p[3]
            if (expression): args.append(expression)
            p[0] = args
        elif (len(p) == 2):
            expression: ast.Expression = p[1]
            p[0] = list([expression])

    def p_expression_atom(self, p: yacc.YaccProduction) -> None:
        '''expression : name
                      | true
                      | false
                      | null
                      | integer
                      | float
                      | complex
                      | string
                      | tuple
                      | list
                      | dict'''
        p[0] = p[1]

    def p_name(self, p: yacc.YaccProduction) -> None:
        '''name : NAME'''
        p[0] = ast.Name(p[1])

    def p_true(self, p: yacc.YaccProduction) -> None:
        '''true : TRUE'''
        p[0] = ast.Boolean(value=True)

    def p_false(self, p: yacc.YaccProduction) -> None:
        '''false : FALSE'''
        p[0] = ast.Boolean(value=False)

    def p_null(self, p: yacc.YaccProduction) -> None:
        '''null : NULL'''
        p[0] = ast.Null(literal=p[1])

    def p_integer(self, p: yacc.YaccProduction) -> None:
        '''integer : INTEGER'''
        p[0] = ast.Integer(literal=p[1])

    def p_float(self, p: yacc.YaccProduction) -> None:
        '''float : FLOAT'''
        p[0] = ast.Float(literal=p[1])

    def p_complex(self, p: yacc.YaccProduction) -> None:
        '''complex : COMPLEX'''
        p[0] = ast.Complex(literal=p[1])

    def p_string(self, p: yacc.YaccProduction) -> None:
        '''string : STRING'''
        p[0] = ast.String(literal=p[1])

    def p_tuple(self, p: yacc.YaccProduction) -> None:
        '''tuple : LPAR expression COMMA expressions RPAR
                 | LPAR expression COMMA RPAR
                 | LPAR RPAR'''
        if (len(p) == 6):
            expression: ast.Expression = p[2]
            expressions: list[ast.Expression] = p[4]
            p[0] = ast.Tuple(expressions=list([expression] + expressions))
        elif (len(p) == 5):
            expression: ast.Expression = p[2]
            p[0] = ast.Tuple(expressions=list([expression]))
        elif (len(p) == 3):
            p[0] = ast.Tuple(expressions=list())

    def p_list(self, p: yacc.YaccProduction) -> None:
        '''list : LSQB expressions RSQB
                | LSQB RSQB'''
        if (len(p) == 4):
            expressions: list[ast.Expression] = p[2]
            p[0] = ast.List(expressions=expressions)
        elif (len(p) == 3):
            p[0] = ast.List(expressions=list())

    def p_dict(self, p: yacc.YaccProduction) -> None:
        '''dict : LBRACE kvpairs RBRACE
                | LBRACE RBRACE'''
        if (len(p) == 4):
            kvpairs: list[tuple[ast.Expression, ast.Expression]] = p[2]
            p[0] = ast.Dict(kvpairs=kvpairs)
        elif (len(p) == 3):
            p[0] = ast.Dict(kvpairs=list())

    def p_expressions(self, p: yacc.YaccProduction) -> None:
        '''expressions : _expressions COMMA
                       | _expressions'''
        p[0] = p[1] 

    def p__expressions(self, p: yacc.YaccProduction) -> None:
        '''_expressions : _expressions COMMA expression
                        | expression'''
        if (len(p) == 4):
            expressions: list[ast.Expression] = p[1]
            expression: ast.Expression = p[3]
            if (expression): expressions.append(expression)
            p[0] = expressions
        elif (len(p) == 2):
            expression: ast.Expression = p[1]
            p[0] = list([expression])

    def p_kvpairs(self, p: yacc.YaccProduction) -> None:
        '''kvpairs : _kvpairs COMMA
                   | _kvpairs'''
        p[0] = p[1]

    def p__kvpairs(self, p: yacc.YaccProduction) -> None:
        '''_kvpairs : _kvpairs COMMA kvpair
                    | kvpair'''
        if (len(p) == 4):
            kvpairs: list[tuple[ast.Expression, ast.Expression]] = p[1]
            kvpair: tuple[ast.Expression, ast.Expression] = p[3]
            if (kvpair): kvpairs.append(kvpair)
            p[0] = kvpairs
        elif (len(p) == 2):
            kvpair: tuple[ast.Expression, ast.Expression] = p[1]
            p[0] = list([kvpair])

    def p_kvpair(self, p: yacc.YaccProduction) -> None:
        '''kvpair : expression COLON expression'''
        key: ast.Expression = p[1]
        value: ast.Expression = p[3]
        p[0] = tuple((key, value,))

    def p_empty(self, p: yacc.YaccProduction) -> None:
        '''empty : '''
        pass

    def p_error(self, p: yacc.YaccProduction) -> None:
        system.GLOBAL.log.debug(f'Error: {p}')
        pass



import sys
import abc
import typing


__all__: list[str] = [
    'STREAM', 'CLASS', 'FUNCTION',
    'Boolean', 'Null', 'Integer', 'Float', 'Complex', 'String', 
    'Tuple', 'List', 'Dict', 'Function', 
    'Annotation', 'Class', 'Property', 'Method', 'Instance',
]


# Anchor stream name to Python stream
STREAM: dict[str, typing.TextIO] = {
    'stdin': sys.stdin,
    'stdout': sys.stdout,
    'stderr': sys.stderr,
}


class Type(abc.ABC):
    
    def __init__(self, typename: str, **kwargs):
        self.__typename: str = typename
        self.__kwargs: dict[str, typing.Any] = kwargs

    @property
    def typename(self) -> str:
        return self.__typename

    @property
    def kwargs(self) -> dict[str, typing.Any]:
        return self.__kwargs


class Boolean(Type, int):

    def __init__(self, value: int, **kwargs):
        Type.__init__(self, 'Boolean', **kwargs)
        self.__value: int = value

    @property
    def value(self) -> int:
        return self.__value


class Null(Type):

    def __init__(self, value: str, **kwargs):
        Type.__init__(self, 'Null', **kwargs)
        self.__value: str = value

    @property
    def value(self) -> str:
        return self.__value

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value


class Integer(Type, int):
    
    def __init__(self, value: int, **kwargs):
        Type.__init__(self, 'Integer', **kwargs)
        self.__value: int = value

    @property
    def value(self) -> int:
        return self.__value


class Float(Type, float):

    def __init__(self, value: float, **kwargs):
        Type.__init__(self, 'Float', **kwargs)
        self.__value: float = value

    @property
    def value(self) -> float:
        return self.__value


class Complex(Type, complex):
    
    def __init__(self, value: complex, **kwargs):
        Type.__init__(self, 'Complex', **kwargs)
        self.__value: complex = value

    @property
    def value(self) -> complex:
        return self.__value


class String(Type, str):
    
    def __init__(self, value: str, **kwargs):
        Type.__init__(self, 'String', **kwargs)
        self.__value: str = value

    @property
    def value(self) -> str:
        return self.__value


class Tuple(Type, tuple): 
    
    def __init__(self, value: tuple, **kwargs):
        Type.__init__(self, 'Tuple', **kwargs)

    def __new__(self, value):
        return tuple.__new__(self, value)


class List(Type, list):

    def __init__(self, value: list, **kwargs):
        Type.__init__(self, 'List', **kwargs)
        self.extend(value)


class Dict(Type, dict):

    def __init__(self, value: dict, **kwargs):
        Type.__init__(self, 'Dict', **kwargs)
        self.update(value)


class Function(Type):

    def __init__(self, **kwargs):
        Type.__init__(self, 'Function', **kwargs)


class Annotation(Type):

    def __init__(self, value: str, **kwargs):
        super().__init__('Annotation', **kwargs)
        self.__value: str = value

    @property
    def value(self) -> str:
        return self.__value


class Class(Type):

    def __init__(self, **kwargs):
        Type.__init__(self, 'Class', **kwargs)


class Property(Type):

    def __init__(self, **kwargs):
        Type.__init__(self, 'Property', **kwargs)


class Method(Type):

    def __init__(self, **kwargs):
        Type.__init__(self, 'Method', **kwargs)


class Instance(Type, object):

    def __init__(self, cls: Class, **kwargs):
        Type.__init__(self, cls.typename, **kwargs)


CLASS = {
    'Boolean': Boolean,
    'Null': Null,
    'Integer': Integer,
    'Float': Float,
    'Complex': Complex,
    'String': String,
    'Tuple': Tuple,
    'List': List,
    'Dict': Dict,
}


def builtin_print(value):
    return print(value, file=STREAM['stdout'])


# Anchor builtin function name to Python function pointer
FUNCTION = {
    'print': builtin_print,
}



import inspect
import typing
import anchor.system as system
import anchor.parse as parse
import anchor.ast as ast
import anchor.symtable as symtable
import anchor.builtins as builtins
import anchor.factory as factory


__all__: list[str] = ['execute',]


def execute(data: str) -> typing.Any:
    # Define main symbol table
    mainidentifier = 'Main'
    symboltable: symtable.SymbolTable = factory.SYMTABLE.new(
        symtable.Type.MAIN, identifier=mainidentifier,
    )

    # Include builtin functions
    for identifier, functionpointer in builtins.FUNCTION.items():
        name = ast.Name(identifier)
        parameters: list[ast.Parameter] = list([
            ast.Parameter(ast.Name(argument))
            for argument in inspect.getfullargspec(functionpointer)[0]
        ])
        functiondef: ast.FunctionDef = ast.FunctionDef(
            name, parameters, None, pointer=functionpointer, isbuiltin=True,
        )
        functiondef.evaluate(symboltable)
    
    # Parse and evaluate abstract syntax tree
    parser: parse.AnchorParser = parse.AnchorParser(
        debuglex=system.GLOBAL.debuglex, 
        debugyacc=system.GLOBAL.debugyacc,
        debuglog=system.GLOBAL.log,
    )
    abstractsyntaxtree: ast.ASTNode = parser.parse(data)
    return abstractsyntaxtree.evaluate(symboltable)



import sys
import typing
import logging


__all__: list[str] = ['GLOBAL',]


class Global(object):

    class __Global(object):

        def __init__(self):
            self.__debug: bool = False
            self.__debuglex: bool = False
            self.__debugyacc: bool = False
            self.__inputstream: typing.TextIO = sys.stdin
            self.__outputstream: typing.TextIO = sys.stdout
            self.__errorstream: typing.TextIO = sys.stderr
            self.__logstream: typing.TextIO = sys.stderr
            self.__log: logging.Logger = None

        @property
        def debug(self) -> bool:
            return self.__debug

        @debug.setter
        def debug(self, other: bool):
            self.__debug = other

        @property
        def debuglex(self) -> bool:
            return self.__debuglex

        @debuglex.setter
        def debuglex(self, other: bool):
            self.__debuglex = other

        @property
        def debugyacc(self) -> bool:
            return self.__debugyacc

        @debugyacc.setter
        def debugyacc(self, other: bool):
            self.__debugyacc = other

        @property
        def inputstream(self) -> typing.TextIO:
            return self.__inputstream

        @inputstream.setter
        def inputstream(self, other: typing.TextIO):
            self.__inputstream = other
            
        @property
        def outputstream(self) -> typing.TextIO:
            return self.__outputstream

        @outputstream.setter
        def outputstream(self, other: typing.TextIO):
            self.__outputstream = other

        @property
        def errorstream(self) -> typing.TextIO:
            return self.__errorstream

        @errorstream.setter
        def errorstream(self, other: typing.TextIO):
            self.__errorstream = other

        @property
        def logstream(self) -> typing.TextIO:
            return self.__logstream

        @logstream.setter
        def logstream(self, other: typing.TextIO):
            self.__logstream = other

        @property
        def log(self) -> logging.Logger:
            return self.__log

        @log.setter
        def log(self, other: logging.Logger):
            self.__log = other

    __instance: __Global = None

    def __new__(cls) -> __Global:
        if (not Global.__instance):
            Global.__instance = Global.__Global()
        return Global.__instance

GLOBAL: Global = Global()



import sys
import os
import typing
import time
import logging
import anchor.system as system
import anchor.compile as compile
import anchor.builtins as builtins


__all__: list[str] = ['main',]


def default(str) -> str:
    return str + ' [Default: %default]'

def readcommand(argv: list[str]) -> dict[str, typing.Any]:
    from optparse import OptionParser
    usage_str = 'anchor [option] [file]'
    parser = OptionParser(usage_str)

    parser.add_option(
        '--debug', action='store_true', dest='debug',
        help=default('debug'), default=False
    )

    parser.add_option(
        '--debug-lex', action='store_true', dest='debuglex',
        help=default('debug lex'), default=False
    )

    parser.add_option(
        '--debug-yacc', action='store_true', dest='debugyacc',
        help=default('debug yacc'), default=False
    )
    
    parser.add_option(
        '--input-stream', dest='inputstream',
        help=default('input stream'), default='stdin'
    )

    parser.add_option(
        '--output-stream', dest='outputstream',
        help=default('output stream'), default='stdout'
    )

    parser.add_option(
        '--error-stream', dest='errorstream',
        help=default('error stream'), default='stderr'
    )

    parser.add_option(
        '--log-stream', dest='logstream',
        help=default('log stream'), 
        default='stderr'
    )

    parser.add_option(
        '--log-file', action='store_true', dest='logfile',
        help=default('generate log file'), default=False
    )

    options, other = parser.parse_args(argv[1:])
    if (len(other) == 0):
        parser.error('no input file')
    if (len(other) > 1):
        parser.error('Command line input not understood: ' + str(other[1:]))

    args = dict()
    args['file'] = other[0]

    # Debug and log stream
    if (options.debug or options.debuglex or options.debugyacc):
        # Debug flags
        system.GLOBAL.debug = options.debug
        if (options.debuglex or options.debugyacc):
            system.GLOBAL.debug = True
            system.GLOBAL.debuglex = options.debuglex
            system.GLOBAL.debugyacc = options.debugyacc

    # Log stream
    logstream = None
    if (options.logfile):
        localtime = time.strftime("%Y%m%d%H%M%S", time.localtime())
        logstream_path = f'./log/anchor_{localtime}.log'
        os.makedirs(os.path.dirname(logstream_path), exist_ok=True)
        logstream = open(logstream_path, 'w')
    elif (options.logstream in builtins.STREAM):
        logstream = builtins.STREAM[options.logstream]
    else:
        logstream = open(options.logstream, 'w')
    system.GLOBAL.logstream = logstream

    # Set up a logging object
    logging.basicConfig(
        level = logging.DEBUG,
        stream = system.GLOBAL.logstream,
        format = '[DEBUG] %(filename)s:%(lineno)d:%(message)s'
    )
    log = logging.getLogger()
    system.GLOBAL.log = log

    # Input stream
    inputstream = None
    if (options.inputstream in builtins.STREAM):
        inputstream = builtins.STREAM[options.inputstream]
    else:
        inputstream = open(options.inputstream, 'r')
    system.GLOBAL.inputstream = inputstream

    # Output stream
    outputstream = None
    if (options.outputstream in builtins.STREAM):
        outputstream = builtins.STREAM[options.outputstream]
    else:
        outputstream = open(options.outputstream, 'w')
    system.GLOBAL.outputstream = outputstream

    # Error stream
    errorstream = None
    if (options.errorstream in builtins.STREAM):
        errorstream = builtins.STREAM[options.errorstream]
    else:
        errorstream = open(options.errorstream, 'w')
    system.GLOBAL.errorstream = errorstream

    return args

def readfile(file) -> str:
    f = open(file, "r", encoding='utf-8')
    data = f.read()
    f.close()
    return data

def main() -> typing.Any:
    args = readcommand(sys.argv)
    data = readfile(**args)
    return compile.execute(data)

main()



import abc
import types
import typing
import anchor.builtins as builtins
import anchor.symtable as symtable
import anchor.factory as factory


__all__: list[str] = []


class ASTNode(abc.ABC):

    @abc.abstractmethod
    def evaluate(self, st): pass


class Atom(abc.ABC):

    __value: builtins.Type = None

    @property
    def value(self):
        return self.__value


class Iterable(abc.ABC):

    @abc.abstractmethod
    def __iter__(self): pass

    @abc.abstractmethod
    def __next__(self): pass

    @abc.abstractmethod
    def __getitem__(self, key): pass


class Callable(abc.ABC):

    @abc.abstractmethod
    def call(self, arguments, st): pass


class Expression(ASTNode):

    @abc.abstractmethod
    def evaluate(self, st) -> ASTNode: pass


class Statement(ASTNode):

    @abc.abstractmethod
    def evaluate(self, st) -> ASTNode: pass


class Block(ASTNode):

    def __init__(self, statements: list[Statement]):
        self.__statements: list[Statement] = statements

    @property
    def statements(self) -> list[Statement]:
        return self.__statements

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        for statement in self.statements:
            astnode: ASTNode = statement.evaluate(st)
            if (isinstance(astnode, Return)):
                return astnode
            elif (isinstance(astnode, Break)):
                return astnode
            elif (isinstance(astnode, Continue)):
                return astnode
        return None


class Program(ASTNode):

    def __init__(self, block: Block):
        self.__block: Block = block

    @property
    def block(self) -> Block:
        return self.__block
    
    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        astnode: ASTNode = self.block.evaluate(st)
        return astnode


class Name(Expression):

    def __init__(self, identifier: str):
        self.__identifier: str = identifier

    @property
    def identifier(self) -> str:
        return self.__identifier

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        symbol: symtable.Symbol = st.lookup(self.identifier)
        astnode: ASTNode = symbol.astnode
        return astnode


class Assignment(Statement):

    def __init__(self, name: Name, expression: Expression):
        self.__name: Name = name
        self.__expression: Expression = expression

    @property
    def name(self) -> Name:
        return self.__name

    @property
    def expression(self) -> Expression:
        return self.__expression

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        identifier: str = self.name.identifier
        astnode = self.expression.evaluate(st)
        astnodes: list[ASTNode] = list([astnode])
        st.insert(identifier, astnodes)
        return None


class Break(Statement):

    def __init__(self, literal: str):
        self.__literal: str = literal

    @property
    def literal(self) -> str:
        return self.__literal

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        st
        return self


class Continue(Statement):
    
    def __init__(self, literal: str):
        self.__literal: str = literal

    @property
    def literal(self) -> str:
        return self.__literal

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        st
        return self


class Return(Statement, Atom):

    __expression: Expression = None
    __value: builtins.Type = None

    def __init__(self, expression: Expression = None, value: builtins.Type = None):
        self.__expression = expression
        self.__value = value

    @property
    def expression(self) -> Expression:
        return self.__expression

    @property
    def value(self) -> builtins.Type:
        return self.__value

    def copy(self) -> ASTNode:
        return Return(value=self.value)

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        if (self.expression != None):
            atom: Atom = self.expression.evaluate(st)
            self.__value = atom.value
        return self.copy()


class Elif(Statement):

    def __init__(self, 
        expression: Expression, block: Block, elseblock: Block = None
    ):
        self.__expression: Expression = expression
        self.__block: Block = block
        self.__elseblock: Block = elseblock

    @property
    def expression(self) -> Expression:
        return self.__expression
    
    @property
    def block(self) -> Block:
        return self.__block

    @property
    def elseblock(self) -> Block:
        return self.__elseblock

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        astnode: ASTNode = self.block.evaluate(st)
        return astnode


class If(Statement):

    def __init__(
        self, expression: Expression, block: Block, 
        elifs: list[Elif] = list(), elseblock: Block = None
    ):
        self.__expression: Expression = expression
        self.__block: Block = block
        self.__elifs: list[Elif] = elifs
        self.__elseblock: Block = elseblock

    @property
    def expression(self) -> Expression:
        return self.__expression
    
    @property
    def block(self) -> Block:
        return self.__block

    @property
    def elifs(self) -> list[Elif]:
        return self.__elifs

    @property
    def elseblock(self) -> Block:
        return self.__elseblock

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        ifcondition: Atom = self.expression.evaluate(st)
        if (ifcondition.value):
            astnode: ASTNode = self.block.evaluate(st)
            return astnode
        else:
            for elifstatement in self.elifs:
                elifcondition: Atom = elifstatement.expression.evaluate(st)
                if (elifcondition.value):
                    astnode: ASTNode = elifstatement.evaluate(st)
                    return astnode
                elif (elifstatement.elseblock):
                    astnode: ASTNode = elifstatement.elseblock.evaluate(st)
                    return astnode
            
            if (self.elseblock):
                astnode: ASTNode = self.elseblock.evaluate(st)
                return astnode
            return None


class Iterate(Statement):

    def __init__(self, iterable: Expression, variable: Name, block: Block):
        self.__iterable: Expression = iterable
        self.__variable: Name = variable
        self.__block: Block = block

    @property
    def iterable(self) -> Expression:
        return self.__iterable
    
    @property
    def variable(self) -> Name:
        return self.__variable

    @property
    def block(self) -> Block:
        return self.__block

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        identifier: str = self.variable.identifier
        for item in self.iterable.evaluate(st):
            itemnode: ASTNode = item
            astnode: ASTNode = itemnode.evaluate(st)
            astnodes: list[ASTNode] = list([astnode])
            st.insert(identifier, astnodes)
            astnode: ASTNode = self.block.evaluate(st)
            if (isinstance(astnode, Return)):
                return astnode
            elif (isinstance(astnode, Break)):
                break
            elif (isinstance(astnode, Continue)):
                continue
        return None


class Loop(Statement):

    def __init__(self, expression: Expression, block: Block):
        self.__expression: Expression = expression
        self.__block: Block = block

    @property
    def expression(self) -> Expression:
        return self.__expression
    
    @property
    def block(self) -> Block:
        return self.__block

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        condition: Atom = self.expression.evaluate(st)
        while (condition.value):
            astnode: ASTNode = self.block.evaluate(st)
            condition = self.expression.evaluate(st)
            if (isinstance(astnode, Return)):
                return astnode
            elif (isinstance(astnode, Break)):
                break
            elif (isinstance(astnode, Continue)):
                continue
        return None


class Annotation(Statement, Atom):

    __value: builtins.Annotation = None

    def __init__(self, literal: typing.Any):
        self.__literal: typing.Any = literal

    @property
    def literal(self) -> typing.Any:
        return self.__literal

    @property
    def value(self) -> builtins.Annotation:
        return self.__value

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        st
        self.__value = builtins.Annotation(str(self.literal))
        return self


class Parameter(Statement):

    def __init__(self, name: Name, typename: Name = None):
        self.__name: Name = name
        self.__typename: Name = typename

    @property
    def name(self) -> Name:
        return self.__name

    @property
    def typename(self) -> Name:
        return self.__typename

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        symbol: symtable.Symbol = st.lookup(self.name.identifier)
        astnode: ASTNode = symbol.astnode
        return astnode


class FunctionDef(Statement, Atom, Callable):

    __value: builtins.Function = None

    def __init__(
        self, name: Name, parameters: list[Parameter], block: Block, **kwargs
    ):
        self.__name: Name = name
        self.__parameters: list[Parameter] = parameters
        self.__block: Block = block
        self.__kwargs: dict[str, typing.Any] = kwargs

    @property
    def name(self) -> Name:
        return self.__name

    @property
    def parameters(self) -> list[Parameter]:
        return self.__parameters

    @property
    def block(self) -> Block:
        return self.__block

    @property
    def kwargs(self) -> dict[str, typing.Any]:
        return self.__kwargs

    @property
    def value(self) -> builtins.Function:
        return self.__value

    def call(
        self, arguments: list[Expression], parentst: symtable.SymbolTable
    ) -> ASTNode:
        parameters: list[Parameter] = self.parameters
        block: Block = self.block
        functionst: symtable.Function = factory.SYMTABLE.new(
            symtable.Type.FUNCTION, 
            identifier=self.name.identifier, parent=parentst,
        )

        # Insert symbols for arguments
        for index in range(len(parameters)):
            parameter: Parameter = parameters[index]
            identifier: str = parameter.name.identifier
            argument: Expression = arguments[index]
            astnode: ASTNode = argument.evaluate(parentst)
            astnodes: list[ASTNode] = list([astnode])
            functionst.insert(identifier, astnodes, isparameter=True)
        
        # Evaluate function block
        isbuiltin = self.kwargs.get('isbuiltin', False)
        if (isbuiltin):
            functionpointer: types.FunctionType = self.kwargs.get('pointer')
            args: dict[str, builtins.Type] = dict()
            for parameter in parameters:
                value: builtins.Type = parameter.evaluate(functionst).value
                args[parameter.name.identifier] = value
            returnvalue: typing.Any = functionpointer(**args)
            return factory.AST.new(value=returnvalue)
        else:
            astnode: ASTNode = block.evaluate(functionst)
            return astnode

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        self.__value = builtins.Function()
        identifier: str = self.name.identifier
        astnodes: list[ASTNode] = list([self])
        st.insert(identifier, astnodes)
        return self


class Property(Statement, Atom):

    __value: builtins.Property = None

    def __init__(self, name: Name, annotations: list[Annotation]):
        self.__name: Name = name
        self.__annotations: list[Annotation] = annotations

    @property
    def name(self) -> Name:
        return self.__name

    @property
    def annotations(self) -> list[Annotation]:
        return self.__annotations

    @property
    def value(self) -> builtins.Property:
        return self.__value

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        self.__value = builtins.Property()
        identifier: str = self.name.identifier
        astnodes: list[ASTNode] = list([self])
        st.insert(identifier, astnodes)
        return self


class MethodDef(Statement, Atom, Callable):

    __value: builtins.Method = None

    def __init__(
        self, name: Name, parameters: list[Parameter], block: Block,
        annotations: list[Annotation], **kwargs
    ):
        self.__name: Name = name
        self.__parameters: list[Parameter] = parameters
        self.__block: Block = block
        self.__annotations: list[Annotation] = annotations
        self.__kwargs: dict[str, typing.Any] = kwargs

    @property
    def name(self) -> Name:
        return self.__name

    @property
    def parameters(self) -> list[Parameter]:
        return self.__parameters

    @property
    def block(self) -> Block:
        return self.__block

    @property
    def annotations(self) -> list[Annotation]:
        return self.__annotations

    @property
    def kwargs(self) -> dict[str, typing.Any]:
        return self.__kwargs

    @property
    def value(self) -> builtins.Method:
        return self.__value

    def call(
        self, arguments: list[Expression], parentst: symtable.SymbolTable
    ) -> ASTNode:
        parameters: list[Parameter] = self.parameters
        block: Block = self.block
        methodst: symtable.Function = factory.SYMTABLE.new(
            symtable.Type.FUNCTION, 
            identifier=self.name.identifier, parent=parentst,
        )

        # Insert symbols for arguments
        for index in range(len(parameters)):
            parameter: Parameter = parameters[index]
            identifier: str = parameter.name.identifier
            astnodes: list[ASTNode] = list([
                arguments[index].evaluate(parentst)
            ])
            methodst.insert(identifier, astnodes, isparameter=True)
        
        # Evaluate method block
        astnode: ASTNode = block.evaluate(methodst)
        return astnode

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        self.__value = builtins.Method()
        identifier: str = self.name.identifier
        astnodes: list[ASTNode] = list([self])
        st.insert(identifier, astnodes)
        return self


class ClassDef(Statement, Atom, Callable):

    __value: builtins.Class = None

    def __init__(self, name: Name, block: Block, annotations: list[Annotation], **kwargs):
        self.__name: Name = name
        self.__block: Block = block
        self.__annotations: list[Annotation] = annotations
        self.__kwargs: dict[str, typing.Any] = kwargs

        self.__properties: dict[str, Property] = dict()
        self.__methods: dict[str, MethodDef] = dict()
        for statement in self.__block.statements:
            if (isinstance(statement, Property)):
                identifier = statement.name.identifier
                self.__properties[identifier] = statement
            elif (isinstance(statement, MethodDef)):
                identifier = statement.name.identifier
                self.__methods[identifier] = statement

        # default factory
        if (name.identifier not in self.__methods):
            parameters = list()
            block = Block(list())
            self.__methods[name.identifier] = \
                MethodDef(name, parameters, block, list())

    @property
    def name(self) -> Name:
        return self.__name

    @property
    def block(self) -> Block:
        return self.__block

    @property
    def kwargs(self) -> dict[str, typing.Any]:
        return self.__kwargs

    @property
    def properties(self) -> dict[str, Property]:
        return self.__properties

    @property
    def methods(self) -> dict[str, MethodDef]:
        return self.__methods

    @property
    def annotations(self) -> list[Annotation]:
        return self.__annotations

    @property
    def value(self) -> builtins.Class:
        return self.__value

    def call(
        self, arguments: list[Expression], parentst: symtable.SymbolTable
    ) -> ASTNode:
        properties: dict[str, Property] = self.properties
        methods: dict[str, MethodDef] = self.methods
        instancest: symtable.Class = factory.SYMTABLE.new(
            symtable.Type.CLASS, 
            identifier=self.name.identifier, parent=parentst,
        )

        # Insert symbols for properties
        for _, prop in properties.items():
            astnodes: list[ASTNode] = list([prop])
            instancest.insert(
                prop.name.identifier, astnodes, isproperty=True
            )

        # Insert symbols for methods
        for _, method in methods.items():
            astnodes: list[ASTNode] = list([method])
            instancest.insert(
                method.name.identifier, astnodes, ismethod=True
            )

        # Evaluate constructor
        factorymethod: MethodDef = methods[self.name.identifier]
        factorymethod.evaluate(instancest)
        call: Call = Call(Name(self.name.identifier), arguments)
        call.evaluate(instancest)

        # Return class instance
        return Instance(self, instancest)

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        self.__value = builtins.Class()
        identifier: str = self.name.identifier
        astnodes: list[ASTNode] = list([self])
        st.insert(identifier, astnodes)
        return self


class Instance(Statement, Atom):

    __value: builtins.Instance = None

    def __init__(self, classdef: ClassDef, instancest: symtable.Class):
        self.__classdef: ClassDef = classdef
        self.__instancest = instancest

    @property
    def classdef(self) -> ClassDef:
        return self.__classdef

    @property
    def instancest(self) -> symtable.Class:
        return self.__instancest

    @property
    def value(self) -> builtins.Instance:
        return self.__value

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        st
        self.__value = builtins.Instance(self.classdef.value)
        return self


class Or(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value or right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class And(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value and right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class Not(Expression):

    def __init__(self, right: Expression):
        self.__right: Expression = right

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        right: Atom = self.right.evaluate(st)
        value: typing.Any = not right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class EqEqual(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value == right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class NotEqual(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value != right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class Less(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value < right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class LessEqual(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value <= right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class Greater(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value > right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class GreaterEqual(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value >= right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class Plus(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value + right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class Minus(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value - right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class Star(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value * right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class DoubleStar(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value ** right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class Slash(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value / right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class DoubleSlash(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value // right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class Percent(Expression):

    def __init__(self, left: Expression, right: Expression):
        self.__left: Expression = left
        self.__right: Expression = right

    @property
    def left(self) -> Expression:
        return self.__left

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        left: Atom = self.left.evaluate(st)
        right: Atom = self.right.evaluate(st)
        value: typing.Any = left.value % right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class UPlus(Expression):

    def __init__(self, right: Expression):
        self.__right: Expression = right

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        right: Atom = self.right.evaluate(st)
        value: typing.Any = +right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class UMinus(Expression):

    def __init__(self, right: Expression):
        self.__right: Expression = right

    @property
    def right(self) -> Expression:
        return self.__right

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        right: Atom = self.right.evaluate(st)
        value: typing.Any = -right.value
        astnode: Expression = factory.AST.new(value=value)
        return astnode


class Boolean(Expression, Atom):

    def __init__(self, value: bool):
        self.__literal: str = str(value)
        self.__value: builtins.Boolean = builtins.Boolean(bool(value))

    @property
    def literal(self) -> str:
        return self.__literal

    @property
    def value(self) -> builtins.Boolean:
        return self.__value

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        st
        return self


class Null(Expression, Atom):

    __literal: str = None
    __value: builtins.Null = None

    def __init__(self, literal: str = None, value: str = None):
        if (literal != None):
            self.__literal = literal
            self.__value = builtins.Null(str(self.literal))
        elif (value != None):
            self.__literal = str(value)
            self.__value = builtins.Null(str(value))

    @property
    def literal(self) -> str:
        return self.__literal

    @property
    def value(self) -> builtins.Null:
        return self.__value

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        st
        return self


class String(Expression, Atom, Iterable):

    __literal: str = None
    __value: builtins.String = None

    def __init__(self, literal: str = None, value: str = None):
        if (literal != None):
            self.__literal = literal
            self.__value = builtins.String(str(self.literal))
        elif (value != None):
            self.__literal = str(value)
            self.__value = builtins.String(str(value))

    @property
    def literal(self) -> str:
        return self.__literal

    @property
    def value(self) -> builtins.String:
        return self.__value

    def __iter__(self):
        self.iter = self.value.__iter__()
        return self

    def __next__(self):
        item = self.iter.__next__()
        return factory.AST.new(value=item)

    def __getitem__(self, key):
        item = self.value.__getitem__(key)
        return factory.AST.new(value=item)

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        st
        return self


class Integer(Expression, Atom):

    __literal: str = None
    __value: builtins.Integer = None

    def __init__(self, literal: str = None, value: int = None):
        if (literal != None):
            self.__literal = literal
            self.__value = builtins.Integer(int(self.literal))
        elif (value != None):
            self.__literal = str(value)
            self.__value = builtins.Integer(int(value))

    @property
    def literal(self) -> str:
        return self.__literal

    @property
    def value(self) -> builtins.Integer:
        return self.__value

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        st
        return self


class Float(Expression, Atom):

    __literal: str = None
    __value: builtins.Float = None

    def __init__(self, literal: str = None, value: float = None):
        if (literal != None):
            self.__literal = literal
            self.__value = builtins.Float(float(self.literal))
        elif (value != None):
            self.__literal = str(value)
            self.__value = builtins.Float(float(value))

    @property
    def literal(self) -> str:
        return self.__literal

    @property
    def value(self) -> builtins.Float:
        return self.__value

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        st
        return self


class Complex(Expression, Atom):

    __literal: str = None
    __value: builtins.Complex = None

    def __init__(self, literal: str = None, value: complex = None):
        if (literal != None):
            self.__literal = literal
            self.__value = builtins.Complex(complex(self.literal))
        elif (value != None):
            self.__literal = str(value)
            self.__value = builtins.Complex(complex(value))

    @property
    def literal(self) -> str:
        return self.__literal

    @property
    def value(self) -> builtins.Complex:
        return self.__value

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        st
        return self


class Tuple(Expression, Atom, Iterable):

    __expressions: list[Expression] = None
    __value: builtins.Tuple = None

    def __init__(
        self, expressions: list[Expression] = None,
        value: tuple = None 
        
    ):
        if (expressions != None):
            self.__expressions: list[Expression] = expressions
        elif (value != None):
            self.__value = builtins.Tuple(tuple(value))

    @property
    def expressions(self) -> list[Expression]:
        return self.__expressions

    @property
    def value(self) -> builtins.Tuple:
        return self.__value

    def __iter__(self):
        self.iter = self.value.__iter__()
        return self

    def __next__(self):
        item = self.iter.__next__()
        return factory.AST.new(value=item)

    def __getitem__(self, key):
        item = self.value.__getitem__(key)
        return factory.AST.new(value=item)
    
    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        if (self.expressions != None):
            atoms: list[Atom] = list([
                expression.evaluate(st)
                for expression in self.expressions
            ])
            self.__value = builtins.Tuple(tuple(map(
                lambda x: x.value, atoms,
            )))
        elif (self.value != None):
            pass
        return self


class List(Expression, Atom, Iterable):
    
    __expressions: list[Expression] = None
    __value: builtins.List = None

    def __init__(
        self, expressions: list[Expression] = None,
        value: list = None
    ):
        if (expressions != None):
            self.__expressions: list[Expression] = expressions
        elif (value != None):
            self.__value = builtins.List(list(value))

    @property
    def expressions(self) -> list[Expression]:
        return self.__expressions

    @property
    def value(self) -> builtins.List:
        return self.__value

    def __iter__(self):
        self.iter = self.value.__iter__()
        return self

    def __next__(self):
        item = self.iter.__next__()
        return factory.AST.new(value=item)

    def __getitem__(self, key):
        item = self.value.__getitem__(key)
        return factory.AST.new(value=item)
    
    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        if (self.expressions != None):
            atoms: list[Atom] = list([
                expression.evaluate(st)
                for expression in self.expressions
            ])
            self.__value = builtins.List(list(map(
                lambda x: x.value, atoms,
            )))
        elif (self.value != None):
            pass
        return self


class Dict(Expression, Atom, Iterable):

    __kvpairs: list[tuple[Expression, Expression]] = None
    __value: builtins.Dict = None

    def __init__(
        self, kvpairs: list[tuple[Expression, Expression]] = None,
        value: dict = None
    ):
        if (kvpairs != None):
            self.__kvpairs: list[tuple[Expression, Expression]] = kvpairs
        elif (value != None):
            self.__value = builtins.Dict(dict(value))

    @property
    def kvpairs(self) -> list[tuple[Expression, Expression]]:
        return self.__kvpairs

    @property
    def value(self) -> builtins.Dict:
        return self.__value

    def __iter__(self):
        self.iter = self.value.__iter__()
        return self

    def __next__(self):
        item = self.iter.__next__()
        return factory.AST.new(value=item)

    def __getitem__(self, key):
        item = self.value.__getitem__(key)
        return factory.AST.new(value=item)

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        if (self.kvpairs != None):
            atoms: dict[Atom, Atom] = dict({
                k.evaluate(st): v.evaluate(st)
                for k, v in self.kvpairs
            })
            self.__value = builtins.Dict(dict(map(
                lambda item: (item[0].value, item[1].value), atoms.items(),
            )))
        elif (self.value != None):
            pass
        return self


class DotName(Expression):

    def __init__(self, expression: Expression, name: Name):
        self.__expression: Expression = expression
        self.__name: Name = name

    @property
    def expression(self) -> Expression:
        return self.__expression

    @property
    def name(self) -> Name:
        return self.__name

    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        st
        return self


class Call(Expression):

    def __init__(self, expression: Expression, arguments: list[Expression]):
        self.__expression: Expression = expression
        self.__arguments: list[Expression] = arguments

    @property
    def expression(self) -> Expression:
        return self.__expression

    @property
    def arguments(self) -> list[Expression]:
        return self.__arguments
    
    def evaluate(self, st: symtable.SymbolTable) -> ASTNode:
        astnode: ASTNode = None
        if (isinstance(self.expression, Name)):
            name = self.expression
            identifier: str = name.identifier
            astnode = st.lookup(identifier).astnode
        elif (isinstance(self.expression, DotName)):
            dotname: DotName = self.expression
            instancename: Name = dotname.expression
            instance: Instance = st.lookup(instancename.identifier).astnode
            st = instance.instancest
            name = dotname.name
            identifier: str = name.identifier
            astnode = st.lookup(identifier).astnode
        else:
            astnode = self.expression.evaluate(st)

        if (isinstance(astnode, Callable)):
            return astnode.call(self.arguments, st)


factory.AST = factory.ASTNodeFactory(declarations=list([
    ('Boolean', Boolean,),
    ('Null', Null,),
    ('Integer', Integer,),
    ('Float', Float,),
    ('Complex', Complex,),
    ('String', String,),
    ('Tuple', Tuple,),
    ('List', List,),
    ('Dict', Dict,),
]))



import anchor.system as system
import anchor.token as token
import anchor.keyword as keyword
import anchor.ply.lex as lex


__all__: list[str] = ['AnchorLexer',]


# Regular expression utility functions
def group(*choices):    return '(' + '|'.join(choices) + ')'
def any(*choices):      return group(*choices) + '*'
def maybe(*choices):    return group(*choices) + '?'

# Comment
Comment                 = r'\#[^\r\n]*'

# Number
Hexnumber               = r'0[xX](?:_?[0-9a-fA-F])+'
Binnumber               = r'0[bB](?:_?[01])+'
Octnumber               = r'0[oO](?:_?[0-7])+'
Decnumber               = r'(?:0(?:_?0)*|[1-9](?:_?[0-9])*)'
Intnumber               = group(Hexnumber, Binnumber, Octnumber, Decnumber)
Exponent                = r'[eE][-+]?[0-9](?:_?[0-9])*'
Pointfloat              = group(
                            r'[0-9](?:_?[0-9])*\.(?:[0-9](?:_?[0-9])*)?',
                            r'\.[0-9](?:_?[0-9])*') + maybe(Exponent
                        )
Expfloat                = r'[0-9](?:_?[0-9])*' + Exponent
Floatnumber             = group(Pointfloat, Expfloat)
Imagnumber              = group(r'[0-9](?:_?[0-9])*[jJ]', Floatnumber + r'[jJ]')
Number                  = group(Imagnumber, Floatnumber, Intnumber)

# String
String = group(r"'[^\n'\\]*(?:\\.[^\n'\\]*)*'", r'"[^\n"\\]*(?:\\.[^\n"\\]*)*"')


class AnchorLexer(object):

    # List of token names
    tokens = tuple(token.NAME.values()) + (
        'NAME', 'INTEGER', 'FLOAT', 'COMPLEX', 'STRING',
    )

    # Regular expression rules for tokens
    t_ignore            = ' \f\t'
    t_ignore_COMMENT    = Comment
    t_INTEGER           = Intnumber
    t_FLOAT             = Floatnumber
    t_COMPLEX           = Imagnumber
    t_STRING            = String
    t_LPAR              = token.pmdict[token.LPAR]
    t_RPAR              = token.pmdict[token.RPAR]
    t_STAR              = token.pmdict[token.STAR]
    t_DOUBLESTAR        = token.pmdict[token.DOUBLESTAR]
    t_PLUS              = token.pmdict[token.PLUS]
    t_COMMA             = token.pmdict[token.COMMA]
    t_MINUS             = token.pmdict[token.MINUS]
    t_RARROW            = token.pmdict[token.RARROW]
    t_DOT               = token.pmdict[token.DOT]
    t_SLASH             = token.pmdict[token.SLASH]
    t_DOUBLESLASH       = token.pmdict[token.DOUBLESLASH]
    t_PERCENT           = token.pmdict[token.PERCENT]
    t_COLON             = token.pmdict[token.COLON]
    t_SEMI              = token.pmdict[token.SEMI]
    t_LESS              = token.pmdict[token.LESS]
    t_LESSEQUAL         = token.pmdict[token.LESSEQUAL]
    t_EQUAL             = token.pmdict[token.EQUAL]
    t_EQEQUAL           = token.pmdict[token.EQEQUAL]
    t_NOTEQUAL          = token.pmdict[token.NOTEQUAL]
    t_GREATER           = token.pmdict[token.GREATER]
    t_GREATEREQUAL      = token.pmdict[token.GREATEREQUAL]
    t_LSQB              = token.pmdict[token.LSQB]
    t_RSQB              = token.pmdict[token.RSQB]
    t_LBRACE            = token.pmdict[token.LBRACE]
    t_RBRACE            = token.pmdict[token.RBRACE]

    # Name
    def t_NAME(self, t: lex.LexToken):
        r'[a-zA-Z_][a-zA-Z0-9_]*'
        if (keyword.iskeyword(t.value)):
            t.type = token.NAME[t.value]
        return t

    # Define a rule for newline so we can track line numbers
    def t_NEWLINE(self, t: lex.LexToken):
        r'\n'
        t.lexer.lineno += 1

    # Error handling rule
    def t_error(self, t: lex.LexToken):
        t.lexer.skip(1)

    # Build the lexer
    def build(self, **kwargs):
        self.lexer = lex.lex(module=self, **kwargs)

    # Test
    def debug(self, data: str) -> None:
        self.lexer.input(data)
        while True:
            t: lex.LexToken = self.lexer.token()
            if (not t): 
                break
            system.GLOBAL.log.debug(t)



import types
import anchor.token as token


__all__: list[str] = ['kwlist', 'iskeyword',]


kwlist: list[str] = list(token.kwdict.values())

iskeyword: types.FunctionType = frozenset(kwlist).__contains__



import typing


__all__: list[str] = ['pmdict', 'kwdict', 'NAME',]


PERCENT: typing.Literal         = 'PERCENT'
LPAR: typing.Literal            = 'LPAR'
RPAR: typing.Literal            = 'RPAR'
STAR: typing.Literal            = 'STAR'
DOUBLESTAR: typing.Literal      = 'DOUBLESTAR'
PLUS: typing.Literal            = 'PLUS'
COMMA: typing.Literal           = 'COMMA'
MINUS: typing.Literal           = 'MINUS'
RARROW: typing.Literal          = 'RARROW'
DOT: typing.Literal             = 'DOT'
SLASH: typing.Literal           = 'SLASH'
DOUBLESLASH: typing.Literal     = 'DOUBLESLASH'
COLON: typing.Literal           = 'COLON'
SEMI: typing.Literal            = 'SEMI'
EQUAL: typing.Literal           = 'EQUAL'
EQEQUAL: typing.Literal         = 'EQEQUAL'
NOTEQUAL: typing.Literal        = 'NOTEQUAL'
LESS: typing.Literal            = 'LESS'
LESSEQUAL: typing.Literal       = 'LESSEQUAL'
GREATER: typing.Literal         = 'GREATER'
GREATEREQUAL: typing.Literal    = 'GREATEREQUAL'
LSQB: typing.Literal            = 'LSQB'
RSQB: typing.Literal            = 'RSQB'
LBRACE: typing.Literal          = 'LBRACE'
RBRACE: typing.Literal          = 'RBRACE'
# tokens defined only for parsing
UPLUS: typing.Literal           = 'UPLUS'
UMINUS: typing.Literal          = 'UMINUS'

pmdict: dict[typing.Literal, str] = {
    COMMA                       : r',',
    LPAR                        : r'\(',
    RPAR                        : r'\)',
    PLUS                        : r'\+',
    MINUS                       : r'-',
    STAR                        : r'\*',
    DOUBLESTAR                  : r'\*\*',
    SLASH                       : r'/',
    DOUBLESLASH                 : r'//',
    PERCENT                     : r'%',
    RARROW                      : r'->',
    DOT                         : r'\.',
    COLON                       : r':',
    SEMI                        : r';',
    EQUAL                       : r'=',
    EQEQUAL                     : r'==',
    NOTEQUAL                    : r'!=',
    LESS                        : r'<',
    LESSEQUAL                   : r'<=',
    GREATER                     : r'>',
    GREATEREQUAL                : r'>=',
    LSQB                        : r'\[',
    RSQB                        : r'\]',
    LBRACE                      : r'{',
    RBRACE                      : r'}',
}


BEGIN: typing.Literal           = 'BEGIN'
END: typing.Literal             = 'END'
TRUE: typing.Literal            = 'TRUE'
FALSE: typing.Literal           = 'FALSE'
NULL: typing.Literal            = 'NULL'
OR: typing.Literal              = 'OR'
AND: typing.Literal             = 'AND'
NOT: typing.Literal             = 'NOT'
CLASS: typing.Literal           = 'CLASS'
PROPERTY: typing.Literal        = 'PROPERTY'
METHOD: typing.Literal          = 'METHOD'
FUNCTION: typing.Literal        = 'FUNCTION'
RETURN: typing.Literal          = 'RETURN'
IF: typing.Literal              = 'IF'
THEN: typing.Literal            = 'THEN'
ELIF: typing.Literal            = 'ELIF'
ELSE: typing.Literal            = 'ELSE'
ITERATE: typing.Literal         = 'ITERATE'
FOR: typing.Literal             = 'FOR'
LOOP: typing.Literal            = 'LOOP'
CONTINUE: typing.Literal        = 'CONTINUE'
BREAK: typing.Literal           = 'BREAK'
PUBLIC: typing.Literal          = 'PUBLIC'
PRIVATE: typing.Literal         = 'PRIVATE'
PROTECTED: typing.Literal       = 'PROTECTED'
FACTORY: typing.Literal         = 'FACTORY'
GET: typing.Literal             = 'GET'
SET: typing.Literal             = 'SET'
REF: typing.Literal             = 'REF'
VAL: typing.Literal             = 'VAL'

kwdict: dict[typing.Literal, str] = {
    BEGIN                       : 'begin',
    END                         : 'end',
    TRUE                        : 'True',
    FALSE                       : 'False',
    NULL                        : 'Null',
    OR                          : 'or',
    AND                         : 'and',
    NOT                         : 'not',
    CLASS                       : 'class',
    PROPERTY                    : 'property',
    METHOD                      : 'method',
    FUNCTION                    : 'function',
    RETURN                      : 'return',
    IF                          : 'if',
    THEN                        : 'then',
    ELIF                        : 'elif',
    ELSE                        : 'else',
    ITERATE                     : 'iterate',
    FOR                         : 'for',
    LOOP                        : 'loop',
    CONTINUE                    : 'continue',
    BREAK                       : 'break',
    PUBLIC                      : 'public',
    PRIVATE                     : 'private',
    PROTECTED                   : 'protected',
    FACTORY                     : 'factory',
    GET                         : 'get',
    SET                         : 'set',
    REF                         : 'ref',
    VAL                         : 'val',
}


__all__.extend(
    [name for name, value in globals().items() if isinstance(value, str)]
)


NAME: dict[typing.Literal, str] = {
    value.replace('\\', ''): name 
    for name, value in (pmdict | kwdict).items()
}



from __future__ import annotations
import typing

if (typing.TYPE_CHECKING):
    import anchor.ast as ast

import anchor.factory as factory


__all__: list[str] = [
    'Type', 'SymbolTable', 'Class', 'Function', 'Symbol',
]


class Symbol(object):

    def __init__(self, identifier: str, astnodes: list[ast.ASTNode], **kwargs):
        self.__identifier: str = identifier
        self.__astnodes: list[ast.ASTNode] = astnodes
        self.__kwargs: dict[str, typing.Any] = kwargs

    @property
    def identifier(self) -> str:
        return self.__identifier

    @property
    def astnodes(self) -> list[ast.ASTNode]:
        return self.__astnodes

    @property
    def astnode(self) -> ast.ASTNode:
        if len(self.__astnodes) != 1:
            # return latest defined astnode
            return self.__astnodes[-1]
        return self.__astnodes[0]

    @property
    def kwargs(self) -> dict[str, typing.Any]:
        return self.__kwargs


class Type(object):
    MAIN = 'MAIN'
    CLASS = 'CLASS'
    FUNCTION = 'FUNCTION'


class SymbolTable(object):

    def __init__(self, identifier: str, parent = None):
        self._identifier: str = identifier
        self._type: Type = Type.MAIN
        self._symbols: dict[str, Symbol] = dict()
        self._parent: SymbolTable = parent

    @property
    def type(self) -> Type:
        return self._type

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def symbols(self) -> dict[str, Symbol]:
        return self._symbols
    
    def insert(
        self, identifier: str, astnodes: list[ast.ASTNode], **kwargs
    ) -> None:
        symbol: Symbol = None
        if (identifier in self.symbols):
            symbol = Symbol(
                identifier, self.symbols[identifier].astnodes + astnodes, 
                **kwargs,
            )
        else:
            symbol = Symbol(identifier, astnodes, **kwargs)
        assert (symbol != None)
        self.symbols[identifier] = symbol

    def lookup(self, identifier: str) -> Symbol:
        if (identifier in self.symbols):
            return self.symbols[identifier]
        elif (self._parent): 
            return self._parent.lookup(identifier)
        else:
            return None


class Class(SymbolTable):

    __methods: list[Symbol] = None

    def __init__(self, identifier: str, parent: SymbolTable = None):
        super().__init__(identifier, parent=parent)
        self._type: Type = Type.CLASS

    @property
    def methods(self) -> list[Symbol]:
        return self.__methods


class Function(SymbolTable):

    __parameters: list[Symbol] = None

    def __init__(self, identifier: str, parent: SymbolTable=None):
        super().__init__(identifier, parent=parent)
        self._type: Type = Type.FUNCTION

    @property
    def parameters(self) -> list[Symbol]:
        return self.__parameters


factory.SYMTABLE = factory.SymbolTableFactory(declarations=list([
    (Type.MAIN, SymbolTable,),
    (Type.CLASS, Class,),
    (Type.FUNCTION, Function,),
]))
